/************************************************************************
 *
 * Constants and vars
 *
 ************************************************************************/
const Cc = Components.classes;
const Ci = Components.interfaces;

//~ extension
const EXT_ID = "imglikeopera@imfo.ru";
const CHROME_CONTENT = "chrome://imglikeopera/content/";

//~ nsIContentPolicy
const REJECT                    = Ci.nsIContentPolicy.REJECT_REQUEST;
const ACCEPT                    = Ci.nsIContentPolicy.ACCEPT;

const TYPE_IMAGE                = Ci.nsIContentPolicy.TYPE_IMAGE;
const TYPE_OBJECT               = Ci.nsIContentPolicy.TYPE_OBJECT;

//~ nsIDOMNode
const TYPE_DOCUMENT_NODE        = Ci.nsIDOMNode.DOCUMENT_NODE;

//~ cache stuff
const nsICache                  = Ci.nsICache;
const nsICacheEntryDescriptor   = Ci.nsICacheEntryDescriptor;

/************************************************************************
 *
 * SubScriptsLoader
 *
 ************************************************************************/
["prefs", "docdata", "file"]
.forEach(
  function(aScriptName) {
    this.loadSubScript(CHROME_CONTENT + "sub-scripts/" + aScriptName + ".js")
  },
  Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader)
)

/************************************************************************
 *
 * Smthng...
 *
 ************************************************************************/

function nsImgLikeOpera() {
  this.debug = false;
  
  this.wrappedJSObject = this;
  
  // cache sessions
  const CACHE_SERVICE = Cc["@mozilla.org/network/cache-service;1"].getService(Ci.nsICacheService);
  
  var httpCacheSession  = CACHE_SERVICE.createSession("HTTP", nsICache.STORE_ANYWHERE, true);
  httpCacheSession.doomEntriesIfExpired = false;
  
  var imageCacheSession = CACHE_SERVICE.createSession("image", nsICache.STORE_ANYWHERE, false);
  imageCacheSession.doomEntriesIfExpired = false;
  
  this.cacheSessions = [httpCacheSession, imageCacheSession];
  
  var observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  observerService.addObserver(this, "profile-after-change", false);
  observerService.addObserver(this, "profile-before-change", false);
  
  this.settings = new PrefsContainer(
    "imglikeopera.", 
    {
      "policy.patterns": "gILO.createPatternsArray",
      "placeholders.font": "gILO.loadUserCss"
    }
  );
  
  this.prefsWrapper = null;
}

/**
 * JS XPCOM component registration
 */
nsImgLikeOpera.prototype = {
  observe: function(aSubject, aTopic, aData) {
    switch (aTopic) {
      case "profile-after-change":
        this.prefsWrapper = new PrefsWrapper();
        
        //~ Set native browser policy: "Load All Images".
        this.prefsWrapper.setIntPref("permissions.default.image", 1);
        
        //~ this.prefsWrapper.setBoolPref("browser.display.show_image_placeholders", true);
        //~ this.prefsWrapper.setBoolPref("browser.display.force_inline_alttext", false);
        
        this.settings.init();
        this.sessionStore.init();
        break;
      
      case "profile-before-change":
        this.sessionStore.uninit();
        this.settings.uninit();
        this.prefsWrapper = null;
        break;
      
      case "app-startup":
        break;
      
      default:
        break;
    }
  },
  
  /**
   * UserCSS
   */
  get userCssFontData() {
    var result = this.settings.placeholdersFont.split(",");
    return result.length == 2 ? result : [];
  },
  
  set userCssFontData(aValue) {
    if (aValue && aValue.length == 2)
      this.settings.placeholdersFont = aValue.join(",");
    
    return this.userCssFontData;
  },
  
  loadUserCss: function() {
    var cssFileContent = ILOFileUtils.readExtensionFile("chrome/skin/user.chrome.css");
    
    var fontSettings = "";
    var [fontFamily, fontSize] = this.userCssFontData;
    
    if (fontFamily)
      fontSettings += "font-family:" + fontFamily + " !important;";
    
    if (fontSize)
      fontSettings += "font-size:" + fontSize + "px !important;";
    
    if (fontSettings)
      cssFileContent = cssFileContent.replace(/\/\*font\*\//, fontSettings);
    
    var sss = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    var ios = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
    var u = ios.newURI("data:text/css," + cssFileContent, null, null);
    
    // Fx2-: not worked, need restart
    if (sss.sheetRegistered(u, sss.USER_SHEET))
      sss.unregisterSheet(u, sss.USER_SHEET);
    
    sss.loadAndRegisterSheet(u, sss.USER_SHEET);
  },
  
  /**
   * StringBundleService helpers
   */
  strBundleManager: {
    get bundle() {
      var bundle = Cc["@mozilla.org/intl/stringbundle;1"]
                   .getService(Ci.nsIStringBundleService)
                   .createBundle("chrome://imglikeopera/locale/imglikeopera.properties");
      delete this.bundle;
      return this.bundle = bundle;
    },
    
    getString: function(aName) {
      return this.bundle.GetStringFromName(aName);
    },
    
    getFormattedString: function(aName, aStrArray) {
      return this.bundle.formatStringFromName(aName, aStrArray, aStrArray.length);
    }
  },
  
  /**
   * SessionStore helpers
   */
  sessionStore: {
    _sessionStoreService: Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore),
    
    _enabled: false,
    
    _checkIsStoreEnabled: function() {
      this._enabled = this._prefsWrapper.getBoolPref("enabled", false);
    },
    
    _prefsWrapper: null,
    
    init: function() {
      this._prefsWrapper = new PrefsWrapper("browser.sessionstore.");
      this._prefsWrapper.branch.addObserver("", this, false);
      this._checkIsStoreEnabled();
    },
    
    uninit: function() {
      this._enabled = false;
      
      if (this._prefsWrapper) {
        this._prefsWrapper.branch.removeObserver("", this);
        this._prefsWrapper = null;
      }
    },
    
    setTabPolicy: function(aTab, aPolicy) {
      if (this._enabled)
        this._sessionStoreService.setTabValue(aTab, "ilo-tab-policy", aPolicy);
    },
    
    getTabPolicy: function(aTab) {
      if (!this._enabled)
        return null;
      
      var policy = parseInt(this._sessionStoreService.getTabValue(aTab, "ilo-tab-policy"), 10);
      return isNaN(policy) ? null : policy;
    },
    
    observe: function(aSubject, aTopic, aData) {
      if (aTopic == "nsPref:changed") {
        switch (aData) {
          case "enabled":
            this._enabled = this._checkIsStoreEnabled();
            break;
          
          default:
            break;
        }
      }
    }
  },
  
  /**
   * Set data container for tab
   */
  setupDocData: function(aTab) {
    return new ILODocData(aTab);
  },
  
  /**
   * Get window (::getWinForObject) and top window (::getTopWinForObject) for document node
   */
  getWinForObject: function(aNode) {
    if (aNode && aNode.nodeType != TYPE_DOCUMENT_NODE)
      aNode = aNode.ownerDocument;
    
    if (!aNode || aNode.nodeType != TYPE_DOCUMENT_NODE)
      return null;
    
    return aNode.defaultView;
  },
  
  getTopWinForObject: function(aNode) {
    var wnd = this.getWinForObject(aNode);
    return wnd ? wnd.top : null;
  },
  
  /**
   * Policy (shouldProcess/shouldLoad)
   */
  shouldProcess: function(contentType, contentLocation, context, obj) {
    return ACCEPT;
  },
  
  /**
   * So... should load?
   */
  shouldLoad: function(contentType, contentLocation, context, obj) {
    if (contentLocation.scheme != "http" && contentLocation.scheme != "https")
      return ACCEPT;
    
    if (contentType != TYPE_IMAGE && (contentType != TYPE_OBJECT || !this.settings.flashBlock))
      return ACCEPT;
    
    var topWnd = this.getTopWinForObject(obj);
    if (!topWnd)
      return ACCEPT;
    
    var tab, doc = topWnd.document;
    
    var docData = doc.iloData;
    
    if (!docData) {
      var docShellTree = topWnd.QueryInterface(Ci.nsIInterfaceRequestor)
                               .getInterface(Ci.nsIWebNavigation)
                               .QueryInterface(Ci.nsIDocShellTreeItem);
      
      if (docShellTree.itemType == Ci.nsIDocShellTreeItem.typeContent) {
        try {
          tab = docShellTree.rootTreeItem
                            .QueryInterface(Ci.nsIInterfaceRequestor)
                            .getInterface(Ci.nsIDOMWindow)
                            .wrappedJSObject
                            .getBrowser()
                            .getBrowserForDocument(doc);
        } catch(e) {}
      }
      
      if (!tab)
        return ACCEPT;
      
      docData = this.setupDocData(tab);
    }
    
    if (!docData || !docData.policy)
      return ACCEPT; //~ smthng wrng
    
    var url = contentLocation.spec;
    
    var docPolicy = docData.policy;
    var fromCss = (contentType == TYPE_IMAGE && obj.nodeType == TYPE_DOCUMENT_NODE);
    
    if (!fromCss) {
      var fixAttr = obj.getAttribute("ilo-ph-fix");
      if (fixAttr)
        return fixAttr == "mustLoad" ?
               this.passObj(obj) : this.blockObj(obj, contentType);
      
      this.changeNodeAttr(obj, "setAttribute", ["ilo-full-src", url]);
    }
    
    if (docData.delayPolicy) {
      if (!fromCss)
        this.changeNodeAttr(obj, "setAttribute", ["ilo-ph-fix", "delay"]);
      
      return REJECT;
    }
    
    if (docPolicy == 3 && !fromCss && this.settings.loadLinkedImages && obj.parentNode.tagName == "A") {
      //~ Load linked images in "Load cached images" mode
      docPolicy = this.settings.loadLinkedImages == 1 ?
                      1 : (this.isThirdPartyHost(doc.location.host, obj.parentNode.host) ? docPolicy : 1);
    } else {
      switch (docData.getURLStatus(url)) {
        case docData.URL_BLOCK:
          return fromCss ? REJECT : this.blockObj(obj, contentType);
        
        case docData.URL_PASS:
          return fromCss ? ACCEPT : this.passObj(obj);
      }
    }
    
    //~ check requested URL in USER filters
    docPolicy = this.getPolicyFromFilters(doc.location, url) || docPolicy;
    
    //~ "Don't load images" => skip work with cache, return false.
    if (docPolicy == 4)
      return fromCss ? REJECT : this.blockObj(obj, contentType);
    
    var expire = this.checkUrlInCache(url, docPolicy);
    var passIt = true;
    
    switch (docPolicy) {
      case 2:
        passIt = expire ? !this.isThirdPartyHost(doc.location.host, contentLocation.host) : true;
        break;
      
      case 3:
        passIt = !expire;
        break;
    }
    
    if (passIt) {
      docData.passURL(url);
      return fromCss ? ACCEPT : this.passObj(obj);
    } else {
      docData.blockURL(url);
      return fromCss ? REJECT : this.blockObj(obj, contentType);
    }
    
    //~ problems? => always return true
    return ACCEPT;
  },
  
  changeNodeAttr: function(aNode, aAction, aActionArguments) {
    if (!(aNode instanceof Ci.nsIDOMNode))
      return;
    
    var wnd = this.getWinForObject(aNode);
    if (!wnd)
      return;
    
    wnd.setTimeout(function() {
      aNode[aAction].apply(aNode, aActionArguments);
    }, 0)
  },
  
  _getObjectWidthAndHeight: function() {
    if (Cc["@mozilla.org/login-manager;1"]) {//Fx3+ (ILOAppInfo.browser.isGreaterThenFx2?)
      return function(aObject) {
        return [
          parseInt(aObject.getAttribute("width") || aObject.style.width, 10),
          parseInt(aObject.getAttribute("height") || aObject.style.height, 10)
        ]
      }
    } else {//Fx2-
      return function(aObject) {
        return [
          parseInt(aObject.width || aObject.style.width, 10),
          parseInt(aObject.height || aObject.style.height, 10)
        ]
      }
    }
  }(),
  
  blockObj: function(obj, contentType) {
    if (obj.hasAttribute("ilo-ph-fix"))
      return REJECT;
    
    var fixAttr = "fixed";
    
    if (contentType == TYPE_OBJECT) {
      fixAttr = "active";
    }
    else if (obj.tagName.toLowerCase() != "input") {
      var [oWidth, oHeight] = this._getObjectWidthAndHeight(obj);
      
      if (!obj.hasAttribute("alt"))
        this.changeNodeAttr(obj, "setAttribute", ["alt", "[Image]"]);
      
      if (!(oWidth || oHeight)) {
        fixAttr = "tofix";
      }
      else if (oWidth && !oHeight) {
        obj.iloFixSH = true;
        obj.height = parseInt(oWidth / 0.668, 10);
      } else if (!oWidth && oHeight) {
        obj.iloFixSW = true;
        obj.width = parseInt(oHeight * 0.668, 10);
      }
    }
    
    this.changeNodeAttr(obj, "setAttribute", ["ilo-ph-fix", fixAttr]);
    
    return REJECT;
  },
  
  passObj: function(obj) {
    if (obj.hasAttribute("ilo-ph-fix")) {
      this.changeNodeAttr(obj, "removeAttribute", ["ilo-ph-fix"]);
      
      if (obj.iloFixSH) {
        obj.iloFixSH = null;
        this.changeNodeAttr(obj, "removeAttribute", ["height"]);
      } else if (obj.iloFixSW) {
        obj.iloFixSW = null;
        this.changeNodeAttr(obj, "removeAttribute", ["width"]);
      }
    }
    
    return ACCEPT;
  },
  
  isThirdPartyHost: function() {
    var _isThirdPartyHost;
    
    if ("nsIEffectiveTLDService" in Ci) {
      var effectiveTLD = Cc["@mozilla.org/network/effective-tld-service;1"].getService(Ci.nsIEffectiveTLDService);
      _isThirdPartyHost = function(aHostA, aHostB) {
        return effectiveTLD.getBaseDomainFromHost(aHostA) !=
               effectiveTLD.getBaseDomainFromHost(aHostB);
      }
    } else {
      _isThirdPartyHost = function(aHostA, aHostB) {
        return aHostA.replace(/.*?((?:[^.]+\.)?[^.]+\.?)$/, "$1") !=
               aHostB.replace(/.*?((?:[^.]+\.)?[^.]+\.?)$/, "$1");
      }
    }
    
    return function(aHostA, aHostB) {
      if (aHostA == aHostB)
        return false;
      
      if (aHostA && aHostB) {
        try {
          return _isThirdPartyHost(aHostA, aHostB);
        } catch(e) {}
      }
      
      return true;
    }
  }(),
  
  checkUrlInCache: function(url, docpolicy) {
    var expired = true;
    
    var forcedExpTime = this.settings.expirationTime;
    var timenow = parseInt((new Date().getTime()) / 1000 - 1, 10);
    
    var cacheEntryDescriptor;
    
    if (forcedExpTime > 0) {
      for each (var cacheSession in this.cacheSessions) {
        try {
          cacheEntryDescriptor = cacheSession.openCacheEntry(url, nsICacheEntryDescriptor, false);
        } catch(e) {}
        
        if (cacheEntryDescriptor) {
          var expirationTime = cacheEntryDescriptor.lastModified + forcedExpTime;
          
          if (expirationTime < timenow) {// || !cacheEntryDescriptor.dataSize(?)) {
            cacheEntryDescriptor.doom();
          } else {
            expired = false;
            
            if (cacheEntryDescriptor.expirationTime != expirationTime) {
              cacheEntryDescriptor.setExpirationTime(expirationTime);
              cacheEntryDescriptor.markValid();
            }
          }
          
          cacheEntryDescriptor.close();
          
          if (!expired) {
            try {
              cacheSession.openCacheEntry(url, nsICache.ACCESS_READ, false);
            } catch(ex) {
              expired = true;
            }
          }
          
          break;//for each
        }
      }
    
    } else if (docpolicy == 3) {
      for each (var cacheSession in this.cacheSessions) {
        try {
          cacheEntryDescriptor = cacheSession.openCacheEntry(url, nsICacheEntryDescriptor, false);
        } catch(e) {}
        
        if (cacheEntryDescriptor) {
          if (cacheEntryDescriptor.expirationTime > timenow)
            expired = false;
          
          cacheEntryDescriptor.close();
          
          break;//for each
        }
      }
    }
    
    return expired;
  },
  
  policyPatternsArray: [],
  
  getPolicyFromFilters: function(aDocLocation, aURL) {
    var policy;
    
    this.policyPatternsArray.some(function(pattern) {
      return (pattern[0].test(aDocLocation) || pattern[0].test(aURL)) ?
                 (policy = pattern[1], true) :
                 false;
    })
    
    return policy;
  },
  
  createPatternsArray: function(aStr) {
    var result = [],
        str2RegExp = this.str2RegExp;
    
    aStr.split("  ").forEach(function(str) {
      var [pattern, policy] = str.split(" ");
      if (pattern && policy) {
        policy = parseInt(policy[0],10);
        if (policy >= 1 && policy <= 4)
          result.push([str2RegExp(pattern), policy]);
      }
    })
    
    this.policyPatternsArray = result;
  },
  
  str2RegExp: function(aStr) {
    //~ not simple; return power
    if (aStr[0] == "/" && aStr[2]) {
      var mod = "", res = "";
      
      var strEnd = aStr.substr(aStr.length-2, aStr.length);
      
      if (strEnd == "/i" || strEnd == "/I") {
        mod = "i";
        res = aStr.substr(1, aStr.length-3);
      } else if (strEnd[1] == "/") {
        res = aStr.substr(1, aStr.length-2);
      }
      
      try {
        return new RegExp(res, mod);
      } catch(e) {}
    }
    
    //~ else: simple;
    res = aStr.replace(/^[\*]*(.(?:[^\*]|[\*]+[^\*])*)[\*]*$/, "$1")
              .replace(/[\*]+/, "*")
              .replace(/([^\w\*])/g, "\\$1")
              .replace(/\*/g, ".*") + "(?:\\n)?";
    
    return new RegExp(res, 'i');
  },


  QueryInterface: function(iid) {
    if (!iid.equals(Ci.nsIImgLikeOpera) &&
        !iid.equals(Ci.nsISupports) &&
        !iid.equals(Ci.nsISupportsWeakReference) &&
        !iid.equals(Ci.nsIContentPolicy) &&
        !iid.equals(Ci.nsIObserver))
    {
      dump("nsImgLikeOpera: QI unknown interface: " + iid + "\n");
      throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
  },
  
  /**
   * log/dump
   */
  console: Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService),
  
  log: function(msg) {
    if (!this.debug)
      return;
    
    msg = "[ILODebugMessage]: " + msg + "\n";
    this.console.logStringMessage(msg);
    dump(msg);
  },
  
  dump: function(aObject) {
    if (!this.debug)
      return;
    
    var dumpTxt = "Dump properties in Object\r\n===============================================\r\n";
    for (var prop in aObject)
      try { dumpTxt += prop + " :: " + aObject[prop] + "\r\n"; } catch(e) {}
    
    this. log(dumpTxt);
  },
  
  dumpObjectInterfaces: function(aObject) {
    if (!this.debug)
      return;
    
    var result = [];
    if (aObject) {
      for each (var iface in Ci) {
        try {
          aObject.QueryInterface(Ci[iface]);
          result.push(iface);
        } catch(e) {}
      }
    }
    
    this. log(result.length ? ["\r\n=== [ object interfaces ] ===\r\n"] + result.join("\r\n") : "[no interfaces]");
  },
  
  /**
   * Component's module implementation
   */
  module: {
    registerSelf: function (compMgr, fileSpec, location, type) {
      var compReg = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
      compReg.registerFactoryLocation(this.cid,
                                      "nsImgLikeOpera JS component",
                                      this.contractId,
                                      fileSpec, location, type );
      var catman = Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
      catman.addCategoryEntry("app-startup", "nsImgLikeOpera", this.contractId, true, true);
      catman.addCategoryEntry("content-policy", this.contractId, this.contractId, true, true);
    },
    
    unregisterSelf: function(compMgr, fileSpec, location) {
      compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
      compMgr.unregisterFactoryLocation(this.cid, fileSpec);
      var catman = Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
      catman.deleteCategoryEntry("app-startup", this.contractId, true);
      catman.deleteCategoryEntry("content-policy", this.contractId, true);
    },
    
    getClassObject: function (compMgr, cid, iid) {
      if (!cid.equals(this.cid))
        throw Components.results.NS_ERROR_NO_INTERFACE;
      if (!iid.equals(Ci.nsIFactory))
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
      return this.factory;
    },
    
    cid: Components.ID("{9aa46f4f-4dc7-4c06-97af-5035170633fe}"),
    
    contractId: "@mozilla.org/imglikeopera;1",
    
    factory: {
      createInstance: function (outer, iid) {
        if (outer != null)
          throw Components.results.NS_ERROR_NO_AGGREGATION;
        return gILO;
      }
    },
  
    canUnload: function(compMgr) {return true;}
  }
}

// NSGetModule: Return the nsIModule object.
function NSGetModule(compMgr, fileSpec) {
    return nsImgLikeOpera.prototype.module;
}

var gILO = new nsImgLikeOpera();